import os
import re
import time
import traceback
import functools
import contextlib
import logging
import collections
import webbrowser
import subprocess

from .vendor import six
from .vendor.Qt import QtCore

_timer = time.time if os.name == "nt" else time.clock
_lru_cache = {}
_threads = []
_basestring = six.string_types[0]  # For Python 2/3
_log = logging.getLogger(__name__)

USE_THREADING = not bool(os.getenv("ALLZPARK_NOTHREADING"))


@contextlib.contextmanager
def timing():
    t0 = _timer()
    result = type("timing", (object,), {"duration": None})
    try:
        yield result
    finally:
        t1 = _timer()
        result.duration = t1 - t0


def delay(func, delay=50):
    """Postpone `func` by `delay` milliseconds

    This is used to allow Qt to finish rendering prior
    to occupying the main thread. Such as calling some
    CPU-heavy function on a `QPushButton.pressed` event,
    which would normally freeze the GUI without letting
    the button unclick itself, resulting in unexpected
    visual artifacts.

    """

    QtCore.QTimer.singleShot(delay, func)


def async_(func):
    """No-op decorator, used to visually distinguish async_ functions"""
    return func


def cached(func):
    """Cache returnvalue of `func`"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        key = "{func}:{args}:{kwargs}".format(
            func=func.__name__,
            args=", ".join(str(arg) for arg in args),
            kwargs=", ".join(
                "%s=%s" % (key, value)
                for key, value in kwargs.items()
            )
        )

        try:
            value = _lru_cache[key]
        except KeyError:
            value = func(*args, **kwargs)
            _lru_cache[key] = value

        return value
    return wrapper


def windows_taskbar_compat():
    """Enable icon and taskbar grouping for Windows 7+"""

    import ctypes
    ctypes.windll.shell32.\
        SetCurrentProcessExplicitAppUserModelID(
            u"allzpark")


if USE_THREADING:
    def defer(target,
              args=None,
              kwargs=None,
              on_success=lambda object: None,
              on_failure=lambda exception: None):
        """Perform operation in thread with callback

        Arguments:
            target (callable): Method or function to call
            callback (callable, optional): Method or function to call
                once `target` has finished.

        Returns:
            None

        """

        thread = Thread(target, args, kwargs, on_success, on_failure)
        thread.finished.connect(lambda: _threads.remove(thread))
        thread.start()

        # Cache until finished
        # If we didn't do this, Python steps in to garbage
        # collect the thread before having had time to finish,
        # resulting in an exception.
        _threads.append(thread)

        return thread

else:
    # Debug mode, execute "threads" immediately on the main thread
    _log.warning("Threading disabled")

    def defer(target,
              args=None,
              kwargs=None,
              on_success=lambda object: None,
              on_failure=lambda exception: None):
        try:
            result = target(*(args or []), **(kwargs or {}))
        except Exception as e:
            on_failure(e)
        else:
            on_success(result)


class Thread(QtCore.QThread):
    succeeded = QtCore.Signal(object)
    failed = QtCore.Signal(Exception, _basestring)

    def __init__(self,
                 target,
                 args=None,
                 kwargs=None,
                 on_success=None,
                 on_failure=None):
        super(Thread, self).__init__()

        self.args = args or list()
        self.kwargs = kwargs or dict()
        self.target = target
        self.on_success = on_success
        self.on_failure = on_failure

        connection = QtCore.Qt.BlockingQueuedConnection

        if on_success is not None:
            self.succeeded.connect(self.on_success, type=connection)

        if on_failure is not None:
            self.failed.connect(self.on_failure, type=connection)

    def run(self, *args, **kwargs):
        try:
            result = self.target(*self.args, **self.kwargs)

        except Exception as e:
            error = traceback.format_exc()
            return self.failed.emit(e, error)

        else:
            self.succeeded.emit(result)


def iterable(arg):
    return (
        isinstance(arg, collections.Iterable)
        and not isinstance(arg, six.string_types)
    )


def open_file_location(fname):
    if os.path.exists(fname):
        if os.name == "nt":
            subprocess.Popen("explorer /select,%s" % fname)
        else:
            webbrowser.open(os.path.dirname(fname))
    else:
        raise OSError("%s did not exist" % fname)


def normpath(path):
    return os.path.normpath(
        os.path.abspath(path).replace("\\", "/")
    )


def normpaths(*paths):
    return list(map(normpath, paths))


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """Key for use with sorted(key=) and str.sort(key=)

    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)

    """

    return [atoi(c) for c in re.split(r'(\d+)', text)]



"""Version includes the Git revision number

This module separates between deployed and development versions of allzpark.
A development version draws its minor version directly from Git, the total
number of commits on the current branch equals the revision number. Once
deployed, this number is embedded into the Python package.

"""

version = "1.3"

try:
    # Look for serialised version
    from .__version__ import version

except ImportError:
    # Else, we're likely running out of a Git repository
    import os as _os
    import subprocess as _subprocess

    try:
        # If used as a git repository
        _cwd = _os.path.dirname(__file__)
        _patch = int(_subprocess.check_output(
            ["git", "rev-list", "HEAD", "--count"],

            cwd=_cwd,

            # Ensure strings are returned from both Python 2 and 3
            universal_newlines=True,

        ).rstrip())

        # Builds since previous minor version
        _patch -= 323

    except Exception:
        # Otherwise, no big deal
        pass

    else:
        version += ".%s" % _patch



import re
import logging

from collections import OrderedDict as odict
from Qt import QtCore, QtWidgets, QtGui

__version__ = "0.5.3"
_log = logging.getLogger(__name__)
_type = type  # used as argument

try:
    # Python 2
    _basestring = basestring
except NameError:
    _basestring = str


class QArgumentParser(QtWidgets.QWidget):
    """User interface arguments

    Arguments:
        arguments (list, optional): Instances of QArgument
        description (str, optional): Long-form text of what this parser is for
        storage (QSettings, optional): Persistence to disk, providing
            value() and setValue() methods

    """

    changed = QtCore.Signal(QtCore.QObject)  # A QArgument

    def __init__(self,
                 arguments=None,
                 description=None,
                 storage=None,
                 parent=None):
        super(QArgumentParser, self).__init__(parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)

        # Create internal settings
        if storage is True:
            storage = QtCore.QSettings(
                QtCore.QSettings.IniFormat,
                QtCore.QSettings.UserScope,
                __name__, "QArgparse",
            )

        if storage is not None:
            _log.info("Storing settings @ %s" % storage.fileName())

        arguments = arguments or []

        assert hasattr(arguments, "__iter__"), "arguments must be iterable"
        assert isinstance(storage, (type(None), QtCore.QSettings)), (
            "storage must be of type QSettings"
        )

        layout = QtWidgets.QGridLayout(self)
        layout.setRowStretch(999, 1)

        if description:
            layout.addWidget(QtWidgets.QLabel(description), 0, 0, 1, 2)

        self._row = 1
        self._storage = storage
        self._arguments = odict()
        self._resets = dict()
        self._description = description

        for arg in arguments or []:
            self._addArgument(arg)

        self.setStyleSheet(style)

    def setDescription(self, text):
        # (TODO) This won't work.
        self._description.setText(text)

    def addArgument(self, name, type=None, default=None, **kwargs):
        # Infer type from default
        if type is None and default is not None:
            type = _type(default)

        # Default to string
        type = type or str

        Argument = {
            None: String,
            int: Integer,
            float: Float,
            bool: Boolean,
            str: String,
            list: Enum,
            tuple: Enum,
        }.get(type, type)

        arg = Argument(name, default=default, **kwargs)
        self._addArgument(arg)
        return arg

    def _addArgument(self, arg):
        if arg["name"] in self._arguments:
            raise ValueError("Duplicate argument '%s'" % arg["name"])

        if self._storage is not None:
            default = self._storage.value(arg["name"])

            if default:
                if isinstance(arg, Boolean):
                    default = bool({
                        None: QtCore.Qt.Unchecked,

                        0: QtCore.Qt.Unchecked,
                        1: QtCore.Qt.Checked,
                        2: QtCore.Qt.Checked,

                        "0": QtCore.Qt.Unchecked,
                        "1": QtCore.Qt.Checked,
                        "2": QtCore.Qt.Checked,

                        # May be stored as string, if used with IniFormat
                        "false": QtCore.Qt.Unchecked,
                        "true": QtCore.Qt.Checked,
                    }.get(default))

                if isinstance(arg, Number):
                    if isinstance(arg, Float):
                        default = float(default)
                    else:
                        default = int(default)

                arg["default"] = default

        # Argument label and editor widget
        label = (
            QtWidgets.QLabel(arg["label"])
            if arg.label
            else QtWidgets.QLabel()
        )
        widget = arg.create()

        for widget in (label, widget):
            widget.setToolTip(arg["help"])
            widget.setObjectName(arg["name"])  # useful in CSS
            widget.setProperty("type", type(arg).__name__)
            widget.setAttribute(QtCore.Qt.WA_StyledBackground)
            widget.setEnabled(arg["enabled"])

        # Reset btn widget
        reset_container = QtWidgets.QWidget()
        reset_container.setProperty("type", "QArgparse:reset")
        reset = QtWidgets.QPushButton("")  # default
        reset.setToolTip("Reset")
        reset.hide()  # shown on edit

        # Align label on top of row if widget is over two times higher
        height = (lambda w: w.sizeHint().height())
        label_on_top = height(label) * 2 < height(widget)
        alignment = (QtCore.Qt.AlignTop,) if label_on_top else ()

        # Layout
        layout = QtWidgets.QVBoxLayout(reset_container)
        layout.addWidget(reset)
        layout.setContentsMargins(0, 0, 0, 0)

        layout = self.layout()
        layout.addWidget(label, self._row, 0, *alignment)
        layout.addWidget(widget, self._row, 1)
        layout.addWidget(reset_container, self._row, 2, *alignment)
        layout.setColumnStretch(1, 1)

        # Signals
        reset.clicked.connect(lambda: arg.write(arg["default"]))
        arg.changed.connect(lambda: self.on_changed(arg))

        self._row += 1
        self._arguments[arg["name"]] = arg
        self._resets[arg["name"]] = reset

    def clear(self):
        assert self._storage, "Cannot clear without persistent storage"
        self._storage.clear()
        _log.info("Clearing settings @ %s" % self._storage.fileName())

    def find(self, name):
        return self._arguments[name]

    def on_changed(self, arg):
        is_edited = arg.read() != arg["default"]

        reset = self._resets[arg["name"]]
        reset.setVisible(is_edited)

        arg["edited"] = is_edited
        self.changed.emit(arg)

    # Optional PEP08 syntax
    add_argument = addArgument


class QArgument(QtCore.QObject):
    """Base class of argument user interface
    """
    changed = QtCore.Signal()

    # Provide a left-hand side label for this argument
    label = True
    # For defining default value for each argument type
    default = None

    def __init__(self, name, default=None, **kwargs):
        super(QArgument, self).__init__(kwargs.pop("parent", None))

        kwargs["name"] = name
        kwargs["label"] = kwargs.get("label", camel_to_title(name))
        kwargs["default"] = self.default if default is None else default
        kwargs["help"] = kwargs.get("help", "")
        kwargs["read"] = kwargs.get("read")
        kwargs["write"] = kwargs.get("write")
        kwargs["enabled"] = bool(kwargs.get("enabled", True))
        kwargs["edited"] = False

        self._data = kwargs

    def __str__(self):
        return self["name"]

    def __repr__(self):
        return "%s(\"%s\")" % (type(self).__name__, self["name"])

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __eq__(self, other):
        if isinstance(other, _basestring):
            return self["name"] == other
        return super(QArgument, self).__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def create(self):
        return QtWidgets.QWidget()

    def read(self):
        return self._read()

    def write(self, value):
        self._write(value)
        self.changed.emit()


class Boolean(QArgument):
    """Boolean type user interface

    Presented by `QtWidgets.QCheckBox`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (bool, optional): Argument's default value, default None
        enabled (bool, optional): Whether to enable this widget, default True

    """
    def create(self):
        widget = QtWidgets.QCheckBox()
        widget.clicked.connect(self.changed.emit)

        if isinstance(self, Tristate):
            self._read = lambda: widget.checkState()
            state = {
                0: QtCore.Qt.Unchecked,
                1: QtCore.Qt.PartiallyChecked,
                2: QtCore.Qt.Checked,
                "1": QtCore.Qt.PartiallyChecked,
                "0": QtCore.Qt.Unchecked,
                "2": QtCore.Qt.Checked,
            }
        else:
            self._read = lambda: bool(widget.checkState())
            state = {
                None: QtCore.Qt.Unchecked,

                0: QtCore.Qt.Unchecked,
                1: QtCore.Qt.Checked,
                2: QtCore.Qt.Checked,

                "0": QtCore.Qt.Unchecked,
                "1": QtCore.Qt.Checked,
                "2": QtCore.Qt.Checked,

                # May be stored as string, if used with QSettings(..IniFormat)
                "false": QtCore.Qt.Unchecked,
                "true": QtCore.Qt.Checked,
            }

        self._write = lambda value: widget.setCheckState(state[value])
        widget.clicked.connect(self.changed.emit)

        if self["default"] is not None:
            self._write(self["default"])

        return widget

    def read(self):
        return self._read()


class Tristate(QArgument):
    """Not implemented"""


class Number(QArgument):
    """Base class of numeric type user interface"""
    default = 0

    def create(self):
        if isinstance(self, Float):
            widget = QtWidgets.QDoubleSpinBox()
            widget.setMinimum(self._data.get("min", 0.0))
            widget.setMaximum(self._data.get("max", 99.99))
        else:
            widget = QtWidgets.QSpinBox()
            widget.setMinimum(self._data.get("min", 0))
            widget.setMaximum(self._data.get("max", 99))

        widget.editingFinished.connect(self.changed.emit)
        self._read = lambda: widget.value()
        self._write = lambda value: widget.setValue(value)

        if self["default"] != self.default:
            self._write(self["default"])

        return widget


class Integer(Number):
    """Integer type user interface

    A subclass of `qargparse.Number`, presented by `QtWidgets.QSpinBox`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (int, optional): Argument's default value, default 0
        min (int, optional): Argument's minimum value, default 0
        max (int, optional): Argument's maximum value, default 99
        enabled (bool, optional): Whether to enable this widget, default True

    """


class Float(Number):
    """Float type user interface

    A subclass of `qargparse.Number`, presented by `QtWidgets.QDoubleSpinBox`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (float, optional): Argument's default value, default 0.0
        min (float, optional): Argument's minimum value, default 0.0
        max (float, optional): Argument's maximum value, default 99.99
        enabled (bool, optional): Whether to enable this widget, default True

    """


class Range(Number):
    """Range type user interface

    A subclass of `qargparse.Number`, not production ready.

    """


class Double3(QArgument):
    """Double3 type user interface

    Presented by three `QtWidgets.QLineEdit` widget with `QDoubleValidator`
    installed.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (tuple or list, optional): Default (0, 0, 0).
        enabled (bool, optional): Whether to enable this widget, default True

    """
    default = (0, 0, 0)

    def create(self):
        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        x, y, z = (self.child_arg(layout, i) for i in range(3))

        self._read = lambda: (
            float(x.text()), float(y.text()), float(z.text()))
        self._write = lambda value: [
            w.setText(str(float(v))) for w, v in zip([x, y, z], value)]

        if self["default"] != self.default:
            self._write(self["default"])

        return widget

    def child_arg(self, layout, index):
        widget = QtWidgets.QLineEdit()
        widget.setValidator(QtGui.QDoubleValidator())

        default = str(float(self["default"][index]))
        widget.setText(default)

        def focusOutEvent(event):
            if not widget.text():
                widget.setText(default)  # Ensure value exists for `_read`
            QtWidgets.QLineEdit.focusOutEvent(widget, event)
        widget.focusOutEvent = focusOutEvent

        widget.editingFinished.connect(self.changed.emit)
        widget.returnPressed.connect(widget.editingFinished.emit)

        layout.addWidget(widget)

        return widget


class String(QArgument):
    """String type user interface

    Presented by `QtWidgets.QLineEdit`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (str, optional): Argument's default value, default None
        placeholder (str, optional): Placeholder message for the widget
        enabled (bool, optional): Whether to enable this widget, default True

    """
    def __init__(self, *args, **kwargs):
        super(String, self).__init__(*args, **kwargs)
        self._previous = None

    def create(self):
        widget = QtWidgets.QLineEdit()
        widget.editingFinished.connect(self.onEditingFinished)
        widget.returnPressed.connect(widget.editingFinished.emit)
        self._read = lambda: widget.text()
        self._write = lambda value: widget.setText(value)

        if isinstance(self, Info):
            widget.setReadOnly(True)
        widget.setPlaceholderText(self._data.get("placeholder", ""))

        if self["default"] is not None:
            self._write(self["default"])
            self._previous = self["default"]

        return widget

    def onEditingFinished(self):
        current = self._read()

        if current != self._previous:
            self.changed.emit()
            self._previous = current


class Info(String):
    """String type user interface but read-only

    A subclass of `qargparse.String`, presented by `QtWidgets.QLineEdit`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (str, optional): Argument's default value, default None
        enabled (bool, optional): Whether to enable this widget, default True

    """


class Color(String):
    """Color type user interface

    A subclass of `qargparse.String`, not production ready.

    """


class Button(QArgument):
    """Button type user interface

    Presented by `QtWidgets.QPushButton`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (bool, optional): Argument's default value, default None
        enabled (bool, optional): Whether to enable this widget, default True

    """
    label = False

    def create(self):
        widget = QtWidgets.QPushButton(self["label"])
        widget.clicked.connect(self.changed.emit)

        state = [
            QtCore.Qt.Unchecked,
            QtCore.Qt.Checked,
        ]

        if isinstance(self, Toggle):
            widget.setCheckable(True)
            self._read = lambda: widget.checkState()
            self._write = (
                lambda value: widget.setCheckState(state[int(value)])
            )
        else:
            self._read = lambda: "clicked"
            self._write = lambda value: None

        if self["default"] is not None:
            self._write(self["default"])

        return widget


class Toggle(Button):
    """Checkable `Button` type user interface

    Presented by `QtWidgets.QPushButton`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        default (bool, optional): Argument's default value, default None
        enabled (bool, optional): Whether to enable this widget, default True

    """


class InfoList(QArgument):
    """String list type user interface

    Presented by `QtWidgets.QListView`, not production ready.

    """
    def __init__(self, name, **kwargs):
        kwargs["default"] = kwargs.pop("default", ["Empty"])
        super(InfoList, self).__init__(name, **kwargs)

    def create(self):
        class Model(QtCore.QStringListModel):
            def data(self, index, role):
                return super(Model, self).data(index, role)

        model = QtCore.QStringListModel(self["default"])
        widget = QtWidgets.QListView()
        widget.setModel(model)
        widget.setEditTriggers(widget.NoEditTriggers)

        self._read = lambda: model.stringList()
        self._write = lambda value: model.setStringList(value)

        return widget


class Choice(QArgument):
    """Argument user interface for selecting one from list

    Presented by `QtWidgets.QListView`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        items (list, optional): List of strings for select, default `["Empty"]`
        default (str, optional): Default item in `items`, use first of `items`
            if not given.
        enabled (bool, optional): Whether to enable this widget, default True

    """
    def __init__(self, name, **kwargs):
        kwargs["items"] = kwargs.get("items", ["Empty"])
        kwargs["default"] = kwargs.pop("default", kwargs["items"][0])
        super(Choice, self).__init__(name, **kwargs)

    def index(self, value):
        """Return numerical equivalent to self.read()"""
        return self["items"].index(value)

    def create(self):
        def on_changed(selected, deselected):
            try:
                selected = selected.indexes()[0]
            except IndexError:
                # At least one item must be selected at all times
                selected = deselected.indexes()[0]

            value = selected.data(QtCore.Qt.DisplayRole)
            set_current(value)
            self.changed.emit()

        def set_current(current):
            options = model.stringList()

            if current == "Empty":
                index = 0
            else:
                for index, member in enumerate(options):
                    if member == current:
                        break
                else:
                    raise ValueError(
                        "%s not a member of %s" % (current, options)
                    )

            qindex = model.index(index, 0, QtCore.QModelIndex())
            smodel.setCurrentIndex(qindex, smodel.ClearAndSelect)
            self["current"] = options[index]

        def reset(items, default=None):
            items = items or ["Empty"]
            model.setStringList(items)
            set_current(default or items[0])

        model = QtCore.QStringListModel()
        widget = QtWidgets.QListView()
        widget.setModel(model)
        widget.setEditTriggers(widget.NoEditTriggers)
        widget.setSelectionMode(widget.SingleSelection)
        smodel = widget.selectionModel()
        smodel.selectionChanged.connect(on_changed)

        self._read = lambda: self["current"]
        self._write = lambda value: set_current(value)
        self.reset = reset

        reset(self["items"], self["default"])

        return widget


class Separator(QArgument):
    """Visual separator

    Example:

        item1
        item2
        ------------
        item3
        item4

    """

    def create(self):
        widget = QtWidgets.QWidget()

        self._read = lambda: None
        self._write = lambda value: None

        return widget


class Enum(QArgument):
    """Argument user interface for selecting one from dropdown list

    Presented by `QtWidgets.QComboBox`.

    Arguments:
        name (str): The name of argument
        label (str, optional): Display name, convert from `name` if not given
        help (str, optional): Tool tip message of this argument
        items (list, optional): List of strings for select, default `[]`
        default (int, optional): Index of default item, use first of `items`
            if not given.
        enabled (bool, optional): Whether to enable this widget, default True

    """
    def __init__(self, name, **kwargs):
        kwargs["default"] = kwargs.pop("default", 0)
        kwargs["items"] = kwargs.get("items", [])

        assert isinstance(kwargs["items"], (tuple, list)), (
            "items must be list"
        )

        super(Enum, self).__init__(name, **kwargs)

    def create(self):
        widget = QtWidgets.QComboBox()
        widget.addItems(self["items"])
        widget.currentIndexChanged.connect(
            lambda index: self.changed.emit())

        self._read = lambda: widget.currentText()
        self._write = lambda value: widget.setCurrentIndex(value)

        if self["default"] is not None:
            self._write(self["default"])

        return widget


style = """\
QWidget {
    /* Explicitly specify a size, to account for automatic HDPi */
    font-size: 11px;
}

*[type="Button"] {
    text-align:left;
}

*[type="Info"] {
    background: transparent;
    border: none;
}

QLabel[type="Separator"] {
    min-height: 20px;
    text-decoration: underline;
}

QWidget[type="QArgparse:reset"] {
    /* Ensure size fixed */
    max-width: 11px;
    max-height: 11px;
    min-width: 11px;
    min-height: 11px;
    padding-top: 0px;
    padding-bottom: 0px;
    padding-left: 0px;
    padding-right: 0px;
}

"""


def camelToTitle(text):
    """Convert camelCase `text` to Title Case

    Example:
        >>> camelToTitle("mixedCase")
        "Mixed Case"
        >>> camelToTitle("myName")
        "My Name"
        >>> camelToTitle("you")
        "You"
        >>> camelToTitle("You")
        "You"
        >>> camelToTitle("This is That")
        "This Is That"

    """

    return re.sub(
        r"((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))",
        r" \1", text
    ).title()


camel_to_title = camelToTitle


def _demo():
    import sys
    app = QtWidgets.QApplication(sys.argv)

    parser = QArgumentParser()
    parser.setWindowTitle("Demo")
    parser.setMinimumWidth(300)

    parser.add_argument("name", default="Marcus", help="Your name")
    parser.add_argument("age", default=33, help="Your age")
    parser.add_argument("height", default=1.87, help="Your height")
    parser.add_argument("alive", default=True, help="Your state")
    parser.add_argument("class", type=Enum, items=[  # (TODO) Reset not hidden
        "Ranger",
        "Warrior",
        "Sorcerer",
        "Monk",
    ], default=2, help="Your class")

    parser.add_argument("options", type=Separator)
    parser.add_argument("paths", type=InfoList, items=[  # (TODO) Doesn't work
        "Value A",
        "Value B",
        "Some other value",
        "And finally, value C",
    ])
    parser.add_argument("location", type=Double3)

    parser.show()
    app.exec_()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="store_true")
    parser.add_argument("--demo", action="store_true")

    opts = parser.parse_args()

    if opts.demo:
        _demo()

    if opts.version:
        print(__version__)



"""Orchestrates view.py and model.py"""

import os
import time
import errno
import shutil
import logging
import tempfile
import threading
import traceback
import subprocess

from collections import OrderedDict as odict

from .vendor.Qt import QtCore, QtGui
from .vendor import transitions
from . import model, util, allzparkconfig

# Third-party dependencies
from . import _rezapi as rez

# Optional third-party dependencies
try:
    from localz import lib as localz
except ImportError:
    localz = None

log = logging.getLogger(__name__)

Latest = model.Latest  # Enum
NoVersion = model.NoVersion


class State(dict):
    """Transient, persistent and machine for state

    The state is used to keep track of which applications
    is the "current" one, along with managing the current
    "state" such as whether the application is busy loading,
    whether it's ready for user input. It also manages persistent
    data, the kind that is stored until the next time the
    application is launched.

    """

    def __init__(self, ctrl, storage):
        super(State, self).__init__({
            "profileName": storage.value("startupProfile"),
            "appRequest": storage.value("startupApplication"),

            # list or callable, returning list of profile names
            "root": None,

            # Current error, if any
            "error": None,

            # Currently commands applications
            "commands": [],

            # Previously loaded profile Rez packages
            "rezProfiles": {},

            # Currently loaded Rez contexts
            "rezContexts": {},

            # Cache, for performance only
            "rezEnvirons": {},

            "rezApps": odict(),
            "fullCommand": "rez env",
            "serialisationMode": (
                storage.value("serialisationMode") or "used_request"
            ),
        })

        self._ctrl = ctrl
        self._storage = storage

    def store(self, key, value):
        """Write to persistent storage

        Arguments:
            key (str): Name of variable
            value (object): Any datatype

        """

        self._storage.setValue(key, value)

    def retrieve(self, key, default=None):
        """Read from persistent storage

        Arguments:
            key (str): Name of variable

        """

        value = self._storage.value(key)

        if value is None:
            value = default

        # Account for poor serialisation format
        # TODO: Implement a better format
        true = ["2", "1", "true", True, 1, 2]
        false = ["0", "false", False, 0]

        if value in true:
            value = True

        if value in false:
            value = False

        return value

    def on_enter_booting(self):
        self._ctrl.debug("Booting..")

    def on_enter_selectprofile(self):
        pass

    def on_enter_resolving(self):
        pass

    def on_enter_launching(self):
        self._ctrl.debug("Application is being launched..")
        util.delay(self.to_ready, 500)

    def on_enter_noapps(self):
        profile = self["profileName"]
        self._ctrl.debug("No applications were found for %s" % profile)

    def on_enter_loading(self):
        self._ctrl.debug("Loading..")

    def on_enter_ready(self):
        self._ctrl.debug("Ready")


class _State(transitions.State):
    def __init__(self, *args, **kwargs):
        help = kwargs.pop("help", "")
        super(_State, self).__init__(*args, **kwargs)
        self.help = help

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.name == other.name
        return self.name == other

    def __ne__(self, other):
        return not self.__eq__(other)


class _Stream(object):
    def __init__(self, ctrl, stream, level):
        self._ctrl = ctrl
        self._stream = stream
        self._level = level

    def write(self, text):
        self._stream.write(text) if self._stream else None
        self._ctrl.logged.emit(text, self._level)

    def fileno(self):
        return 0

    def close(self):
        return


class Controller(QtCore.QObject):
    state_changed = QtCore.Signal(_State)
    logged = QtCore.Signal(str, int)  # message, level
    resetted = QtCore.Signal()

    # One or more packages have changed on disk
    repository_changed = QtCore.Signal()

    profile_changed = QtCore.Signal(
        str, object, bool)  # profile, version, refreshed

    application_changed = QtCore.Signal()

    # The current command to launch an application has changed
    command_changed = QtCore.Signal(str)  # command

    patch_changed = QtCore.Signal(str)  # full patch string

    states = [
        _State("booting", help="ALLZPARK is booting, hold on"),
        _State("resolving", help="Rez is busy resolving a context"),
        _State("loading", help="Something is taking a moment"),
        _State("errored", help="Something has gone wrong"),
        _State("launching", help="An application is launching"),
        _State("ready", help="Awaiting user input"),
        _State("noprofiles", help="Allzpark did not find any profiles at all"),
        _State("noapps", help="There were no applications to choose from"),
        _State("notresolved", help="Rez couldn't resolve a request"),
        _State("pkgnotfound", help="One or more packages was not found"),
    ]

    def __init__(self, storage, stdio=None, stderr=None, parent=None):
        super(Controller, self).__init__(parent)

        state = State(self, storage)

        models = {
            "apps": model.ApplicationModel(),
            "profileVersions": QtCore.QStringListModel(),

            # Docks
            "profiles": model.ProfileModel(),
            "packages": model.PackagesModel(self),
            "context": model.ContextModel(),
            "environment": model.EnvironmentModel(),
            "commands": model.CommandsModel(),
        }

        timers = {
            "commandsPoller": QtCore.QTimer(self),
        }

        timers["commandsPoller"].timeout.connect(self.on_tasks_polled)
        timers["commandsPoller"].start(500)

        # Initialize the state machine
        self._machine = transitions.Machine(
            model=state,
            states=self.states,
            initial="booting",
            after_state_change=[self.on_state_changed],

            # Add "on_enter_<state name>" to model
            auto_transitions=True,
        )

        self._timers = timers
        self._models = models
        self._storage = storage
        self._state = state
        self._name_to_state = {
            state.name: state
            for state in self.states
        }

        state.on_enter_booting()

    # ----------------
    # Data
    # ----------------

    @property
    def models(self):
        return self._models

    @property
    def state(self):
        return self._state

    @property
    def current_error(self):
        return self._state["error"]

    @property
    def current_profile(self):
        return self._state["profileName"]

    @property
    def current_application(self):
        return self._state["appRequest"]

    @property
    def current_tool(self):
        return self._state["tool"]

    def context(self, app_request):
        return self._state["rezContexts"][app_request].to_dict()

    def environ(self, app_request):
        """Fetch the environment of a context

        NOTE: These can get very expensive. They call on every
              package.py:commands() in a resolved context, which can
              be in the tens to hundreds. Add to that the fact that
              these functions can perform any arbitrary task, including
              writing to disk or performing expensive calculations,
              such as resolving their own contexts for various reasons.

        TODO: This should be async, the GUI should help the user
              understand that the environment is loading and is
              going to be ready soon. They should also only incur
              cost when the user is actually looking at the
              environment tab.

        """

        env = self._state["rezEnvirons"]
        ctx = self._state["rezContexts"]

        try:
            return env[app_request]

        except KeyError:
            try:
                environ = ctx[app_request].get_environ()
                env[app_request] = environ
                return environ

            except rez.ResolvedContextError:
                return {
                    "error": "Failed context"
                }

    def resolved_packages(self, app_request):
        return self._state["rezContexts"][app_request].resolved_packages

    # ----------------
    # Events
    # ----------------

    def on_tasks_polled(self):
        self._models["commands"].poll()

    def on_state_changed(self):
        state = self._name_to_state[self._state.state]
        self.state_changed.emit(state)

    def on_unhandled_exception(self, type, value, tb):
        """From sys.excepthook

        Exceptions are normally handled close to the caller,
        but some exceptions are better handled globally. For
        example, if a PackageCommandError occurs, there is
        little a caller can do. The problem must be addressed
        by the user, outside of the entire program, so best we
        can do is let them know as nicely as possible.

        Arguments:
            type (Exception): Subclass of Exception
            value (str): Message to the user
            tb (str): Full traceback

        Returns:
            handled (bool): True is the application dealt
                with it, False otherwise. An unhandled exception
                is raised to the command-line/caller.

        """

        # Potentially overridden by the below
        self._state["error"] = "".join(
            traceback.format_tb(tb) + [str(value)]
        )
        self._state.to_errored()
        self.error(self._state["error"])

        if rez.PackageNotFoundError is type:
            package = value.value.rsplit(": ", 1)[-1]
            paths = self._package_paths()
            message = """
                <h2><font color=\"red\">:(</font></h2>

                Package '{package}' is required by this profile,
                but could <font color=\"red\">not be found.</font>
                <br>
                <br>
                I searched in these paths:
                {paths}
            """

            self._state["error"] = message.format(
                package=package,
                paths="<ul>%s</ul>" % "".join(
                    "<li>%s</li>" % path for path in paths
                )
            )

            self._state.to_noapps()
            return True

        elif rez.PackageFamilyNotFoundError is type:
            # package family not found: occoc (searched: C:\)
            _, package, paths = value.value.split(": ", 2)
            package = package.split(" (", 1)[0]
            paths = paths.rstrip(")").split("; ")  # Hard-coded pathsep in Rez

            message = """
                <h2><font color=\"red\">:(</font></h2>

                Package '{package}' is required by this profile,
                but could <font color=\"red\">not be found.</font>
                <br>
                <br>
                I searched in these paths:
                {paths}
            """

            self._state["error"] = message.format(
                package=package,
                paths="<ul>%s</ul>" % "".join(
                    "<li>%s</li>" % path for path in paths
                )
            )

            self._state.to_noapps()
            return True

        elif rez.ResolvedContextError is type:
            # Cannot perform operation in a failed context
            self.error(str(value))
            self._state.to_ready()
            return True

        elif rez.RexError is type:
            # These are re-raised as a more specific
            # exception, e.g. RexUndefinedVariableError
            self._state.to_errored()

        elif rez.RexUndefinedVariableError is type:
            self._state.to_errored()

        elif rez.PackageCommandError is type:
            self._state.to_errored()

        elif rez.PackageRequestError is type:
            message = "<h2><font color=\"red\">:(</font></h2>%s"
            self._state["error"] = message % value
            self._state.to_noapps()

        self.error(self._state["error"])

    # ----------------
    # Methods
    # ----------------

    def stdio(self, stream, level=logging.INFO):
        return _Stream(self, stream, level)

    def find(self, family, range_=None):
        """Find packages, relative Allzpark state

        Arguments:
            family (str): Name of package
            range_ (str): Range, e.g. "1" or "==0.3.13"

        """

        package_filter = self._package_filter()
        paths = self._package_paths()
        it = rez.find(family, range_, paths=paths)
        it = sorted(
            it,

            # Make e.g. 1.10 appear after 1.9
            key=lambda p: util.natural_keys(str(p.version))
        )

        for pkg in it:
            if package_filter.excludes(pkg):
                self.debug("Excluding %s==%s.." % (pkg.name, pkg.version))
                continue

            yield pkg

    def env(self, request, use_filter=True):
        """Resolve context, relative Allzpark state

        Arguments:
            request (str): Fully formatted request, including any
                number of packages. E.g. "six==1.2 PySide2"
            use_filter (bool, optional): Whether or not to apply
                the current package_filter

        """

        package_filter = self._package_filter()
        paths = self._package_paths()

        return rez.env(
            request,
            package_paths=paths,
            package_filter=package_filter if use_filter else None
        )

    def update_command(self, mode=None):
        if mode:
            self._state["serialisationMode"] = mode
            self._state.store("serialisationMode", mode)

        if self._state["appRequest"] not in self._state["rezContexts"]:
            # In this case, we have no context, so there
            # is very little to actually try and reproduce
            self._state["fullCommand"] = ""
            return self.command_changed.emit("")

        mode = self._state["serialisationMode"]
        app = self._state["appRequest"]
        context = self._state["rezContexts"][app]
        tool = self._state["tool"]
        exclude = allzparkconfig.exclude_filter

        if mode == "used_resolve":
            packages = [
                "%s==%s" % (pkg.name, pkg.version)
                for pkg in context.resolved_packages or []
            ]

        else:
            packages = [str(pkg) for pkg in context.requested_packages()]

        command = ["rez", "env"]
        command += packages

        if exclude:
            command += ["--exclude", exclude]

        # Ensure consistency during re-resolve
        # Important for submitting contexts across
        # machines at different times
        command += ["--time", str(context.timestamp)]

        if localz and not self._state.retrieve("useLocalizedPackages", True):
            paths = os.pathsep.join(self._package_paths())
            command += ["--paths"] + ["\"%s\"" % paths]

        elif not self._state.retrieve("useDevelopmentPackages"):
            command += ["--no-local"]

        command += ["--", tool]

        self._state["fullCommand"] = " ".join(command)

        self.command_changed.emit(self._state["fullCommand"])

    def _package_filter(self):
        package_filter = rez.PackageFilterList.singleton.copy()

        if allzparkconfig.exclude_filter:
            rule = rez.Rule.parse_rule(allzparkconfig.exclude_filter)
            package_filter.add_exclusion(rule)

        return package_filter

    @util.async_
    def reset(self, root=None, on_success=lambda: None):
        """Initialise controller with `root`

        Profiles are listed at `root` and matched
        with its corresponding Rez package.

        Arguments:
            root (list, callable): A list of profile names, or a callable
                returning names of profiles.
            on_success (callable): Callback on reset completed.

        """

        self.info("Resetting..")
        root = root or self._state["root"]
        assert root, "Tried resetting without a root, this is a bug"

        def do():
            profiles = dict()
            default_profile = None

            for name in self.list_profiles(root):

                # Find profile package
                package = None
                for package in self.find(name):

                    if name not in profiles:
                        profiles[name] = dict()

                    profiles[name][str(package.version)] = package
                    profiles[name][Latest] = package

                if package is None:
                    package = model.BrokenPackage(name)
                    profiles[name] = {
                        "0.0": package,
                        Latest: package,
                    }

                # Default to latest of last
                default_profile = name

            self._state["rezProfiles"].update(profiles)

            # On resetting after startup, there will be a
            # currently selected profile that may differ from
            # the startup profile.
            current_profile = self._state["profileName"]

            if current_profile and current_profile not in profiles:
                self.warning("Startup profile '%s' did not exist"
                             % current_profile)
                current_profile = None

            # The user has never opened the GUI before,
            # or user preferences has been wiped.
            if not current_profile:
                current_profile = default_profile

            self._models["profiles"].set_favorites(self)
            self._models["profiles"].set_current(current_profile)
            self._models["profiles"].reset(profiles)

            self._state["profileName"] = current_profile
            self._state["root"] = root

            self._state.to_ready()
            self.resetted.emit()

        def _on_success():
            profile = not self._state["profileName"]

            if profile:
                self._state.to_noprofiles()
            else:
                self.select_profile(profile)

            on_success()

        def _on_failure(error, trace):
            raise error

        self._state["rezContexts"].clear()
        self._state["rezEnvirons"].clear()
        self._state["rezApps"].clear()

        # Rez stores file listings and more
        # in memory, in addition to memcached.
        # This function clears the in-memory cache,
        # so that we can pick up new packages.
        rez.clear_caches()

        self._state.to_booting()
        util.defer(
            do,
            on_success=_on_success,
            on_failure=_on_failure
        )

    def patch(self, new):
        self.debug("Patching %s.." % new)

        new = rez.PackageRequest(new)
        old = odict(
            (rez.PackageRequest(req).name, rez.PackageRequest(req))
            for req in self._state.retrieve("patch", "").split()
        )

        if new.name in old:
            old.pop(new.name)

        if str(new.range):
            # Otherwise, let it return to the originally resolved value
            old[new.name] = new

        patch = " ".join(str(pkg) for pkg in old.values())
        self._state.store("patch", patch)
        self.reset()

    @util.async_
    def launch(self, **kwargs):
        def do():
            app_request = self._state["appRequest"]
            rez_context = self._state["rezContexts"][app_request]
            rez_app = self._state["rezApps"][app_request]

            self.debug("Found app: %s=%s" % (
                rez_app.name, rez_app.version
            ))

            app_model = self._models["apps"]
            app_index = app_model.findIndex(app_request)

            tool_name = kwargs.get(
                "command", app_model.data(app_index, "tool"))
            is_detached = kwargs.get(
                "detached", app_model.data(app_index, "detached"))

            assert tool_name, (
                "There should have been at least one tool name. "
                "This is a bug"
            )

            overrides = self._models["packages"]._overrides
            disabled = self._models["packages"]._disabled
            environ = self._state.retrieve("userEnv", {})

            self.debug(
                "Launching %s%s.." % (
                    tool_name, " (detached)" if is_detached else "")
            )

            def on_error(error):
                # Forward error from Command()
                raise error

            cmd = Command(
                context=rez_context,
                command=tool_name,
                package=rez_app,
                overrides=overrides,
                disabled=disabled,
                detached=is_detached,
                environ=environ,
                parent=self
            )

            cmd.stdout.connect(self.info)
            cmd.stderr.connect(self.error)
            cmd.error.connect(on_error)

            cmd.execute()

            self._state["commands"].append(cmd)
            self._models["commands"].append(cmd)

            self._state.store("app/%s/lastUsed" % app_request, time.time())
            self._state.to_launching()

        self._state.to_loading()
        util.delay(do)

    def localize(self, name):
        tempdir = tempfile.mkdtemp()

        def do():
            self.debug("Resolving %s.." % name)
            variant = localz.resolve(name)[0]  # Guaranteed to be one

            try:
                self.debug("Preparing %s.." % name)
                copied = localz.prepare(variant, tempdir, verbose=2)[0]

                self.debug("Computing size..")
                size = localz.dirsize(tempdir) / (10.0 ** 6)  # mb

                self.debug("Localising %.2f mb.." % size)
                result = localz.localize(copied,
                                         localz.localized_packages_path(),
                                         verbose=2)

                self.debug("Localised %s" % result)

            finally:
                self.debug("Cleaning up..")
                shutil.rmtree(tempdir)

        def on_success(result):
            self.repository_changed.emit()

        def on_failure(error, trace):
            self.error(trace)

        util.defer(do,
                   on_success=on_success,
                   on_failure=on_failure)

    def delocalize(self, name):
        def do():
            item = self._models["packages"].find(name)
            package = item["package"]
            self.debug("Delocalizing %s" % package.root)
            localz.delocalize(package)

        def on_success(result):
            self.repository_changed.emit()

        def on_failure(error, trace):
            self.error(trace)

        util.defer(do,
                   on_success=on_success,
                   on_failure=on_failure)

    def _localize_status(self, package):
        """Return status of localisation"""
        return None

    def debug(self, message):
        self.logged.emit(message, logging.DEBUG)

    def info(self, message):
        self.logged.emit(message, logging.INFO)

    def warning(self, message):
        self.logged.emit(message, logging.WARNING)

    def error(self, message):
        self.logged.emit(str(message), logging.ERROR)

    def list_profiles(self, root=None):
        root = root or self._state["root"]
        assert root, "Tried listing without a root, this is a bug"

        if isinstance(root, (tuple, list)):
            profiles = root

        elif callable(root):
            try:
                profiles = root()

            except Exception:
                if log.level == logging.DEBUG:
                    traceback.print_exc()

                self.error("Could not find profiles in %s" % root)
                profiles = []

        else:
            raise TypeError("Argument 'root' should be either list type or "
                            "callable.")

        # Facilitate accidental empty family names, e.g. None or ''
        profiles = list(filter(None, profiles))

        return profiles

    @util.async_
    def select_profile(self, profile_name, version_name=Latest):

        # Wipe existing data
        self._models["apps"].reset()
        self._models["context"].reset()
        self._models["environment"].reset()
        self._models["packages"].reset()
        self._models["profileVersions"].setStringList([])

        self._state["rezContexts"].clear()
        self._state["rezEnvirons"].clear()
        self._state["rezApps"].clear()

        def on_apps_found(apps):
            if not apps:
                self._state["error"] = """
                <h2><font color=\"red\">:(</font></h2>
                <br>
                <br>
                The profile was found, but no applications.
                <br>
                <br>
                The profile didn't specify an application for you to use.<br>
                This is likely due to a misconfigured profile. Don't forget<br>
                to provide one or more packages as <i>weak references</i>.
                <br>
                <br>
                See <a href=https://allzpark.com/getting-started>
                    allzpark.com/getting-started</a> for more details.

                """
                self._state.to_noapps()

            else:
                self._models["apps"].reset(apps)
                self._state.to_ready()

        def on_apps_not_found(error, trace):
            # Handled by on_unhandled_exception
            raise error

        try:
            profile_versions = self._state["rezProfiles"][profile_name]
            active_profile = profile_versions[version_name]

        except KeyError:
            # This can only happen if somehow the view decided to pass
            # along the name and version of a profile that didn't exist.
            profile_name = self._state["profileName"]
            profile_versions = self._state["rezProfiles"][profile_name]
            active_profile = profile_versions[version_name]

            if profile_name:
                self.warning("%s was not found" % profile_name)
            else:
                self.error("select_profile was passed an empty string")

        refreshed = self._state["profileName"] == profile_name

        # TODO: This isn't clear.
        # We can't pass a native Rez Version object, but we also can't
        # simply str() that and BrokenPackage.version, as those would
        # be None, which is the equivalent of a NoVersion object.
        version_name = active_profile.version
        version_name = str(version_name) if version_name else NoVersion

        self._state["profileName"] = profile_name
        self.profile_changed.emit(
            profile_name,
            version_name,
            refreshed
        )

        if isinstance(active_profile, model.BrokenPackage):
            raise rez.PackageNotFoundError(
                "package not found: %s" % profile_name
            )

        # Update versions model
        versions = list(filter(None, profile_versions))  # Exclude "Latest"
        versions.reverse()  # Latest first
        self._models["profileVersions"].setStringList(versions)

        self._state.to_loading()
        util.defer(
            self._list_apps,
            args=[active_profile],
            on_success=on_apps_found,
            on_failure=on_apps_not_found,
        )

    def select_application(self, app_request):
        self._state["appRequest"] = app_request

        try:
            context = self.context(app_request)
            environ = self.environ(app_request)
            packages = self.resolved_packages(app_request)

        except Exception:
            self._models["packages"].reset()
            self._models["context"].reset()
            self._models["environment"].reset()
            raise

        self._models["packages"].reset(packages)
        self._models["context"].load(context)
        self._models["environment"].load(environ)

        tools = self._models["apps"].find(app_request)["tools"]
        self._state["tool"] = tools[0]

        # Use this application on next launch or change of profile
        self.update_command()
        self._state.store("startupApplication", app_request)
        self.application_changed.emit()

    def select_tool(self, tool_name):
        self._state["tool"] = tool_name
        self.update_command()

    def _package_paths(self):
        """Return all package paths, relative the current state of the world"""

        paths = rez.config.packages_path[:]

        # Optional development packages
        if not self._state.retrieve("useDevelopmentPackages"):
            paths = rez.config.nonlocal_packages_path[:]

        # Optional package localisation
        if localz and not self._state.retrieve("useLocalizedPackages", True):
            path = localz.localized_packages_path()

            try:
                paths.remove(util.normpath(path))
            except ValueError:
                # It may not be part of the path
                pass

        return paths

    def _list_apps(self, profile):
        # Each app has a unique context relative the current profile
        # Find it, and keep track of it.

        apps = []
        _apps = allzparkconfig.applications

        if self._state.retrieve("showAllApps") and not _apps:
            self.warning("Requires allzparkconfig.applications")

        elif self._state.retrieve("showAllApps"):
            if isinstance(_apps, (tuple, list)):
                apps = _apps

            else:
                try:
                    if callable(_apps):
                        apps = _apps()
                    else:
                        apps = os.listdir(_apps)
                except OSError as e:
                    if e.errno not in (errno.ENOENT,
                                       errno.EEXIST,
                                       errno.ENOTDIR):
                        raise

                    self.warning("Could not show all apps, "
                                 "missing `allzparkconfig.applications`")

        if not apps:
            apps[:] = allzparkconfig.applications_from_package(profile)

        # Optional patch
        patch = self._state.retrieve("patch", "").split()
        package_filter = self._package_filter()

        contexts = odict()
        with util.timing() as t:
            for app_request in apps:
                app_request = rez.PackageRequest(app_request.strip("~"))
                app_package = rez.find_latest(app_request.name,
                                              range_=app_request.range)

                if package_filter.excludes(app_package):
                    continue

                variants = list(profile.iter_variants())
                variant = variants[0]

                if len(variants) > 1:
                    # Unsure of whether this is desirable. It would enable
                    # a profile per platform, or potentially other kinds
                    # of special-purpose situations. If you see this,
                    # and want this, submit an issue with your use case!
                    self.warning(
                        "Profiles with multiple variants are unsupported. "
                        "Using first found: %s" % variant
                    )

                app_request = "%s==%s" % (app_package.name,
                                          app_package.version)

                request = [variant.qualified_package_name, app_request]
                self.debug("Resolving request: %s" % " ".join(request))

                context = self.env(request)

                if context.success and patch:
                    self.debug("Patching request: %s" % " ".join(request))
                    request = context.get_patched_request(patch)
                    context = self.env(
                        request,
                        use_filter=self._state.retrieve(
                            "patchWithFilter", False
                        )
                    )

                contexts[app_request] = context

        # Associate a Rez package with an app
        for app_request, rez_context in contexts.items():
            try:
                rez_pkg = next(
                    pkg
                    for pkg in rez_context.resolved_packages
                    if "%s==%s" % (pkg.name, pkg.version) == app_request
                )

            except StopIteration:
                rez_pkg = model.BrokenPackage(app_request)

                self.warning(
                    "Couldn't find a corresponding package for "
                    "application %s. This can happen if an application is "
                    "patched away, using the ^-operator."
                    % app_request
                )

            except TypeError:
                # resolved_packages was None, a sign that a context was broken
                rez_pkg = model.BrokenPackage(app_request)

                if rez_context.success:
                    self.warning(
                        "This shouldn't have happened, "
                        "I was expecting a broken context here. "
                        "Please report this to "
                        "https://github.com/mottosso/allzpark/issues/66"
                    )

                self.error(
                    "Context for '%s' had no resolved packages, this is "
                    "likely due to a version conflict and broken resolve. "
                    "Try graphing it." % app_request
                )

            self._state["rezApps"][app_request] = rez_pkg

        self.debug("Resolved all contexts in %.2f seconds" % t.duration)

        # Hide hidden
        visible_apps = []
        show_hidden = self._state.retrieve("showHiddenApps")
        for request, package in self._state["rezApps"].items():
            data = allzparkconfig.metadata_from_package(package)
            hidden = data.get("hidden", False)

            if hidden and not show_hidden:
                continue

            visible_apps += [package]

        self._state["rezContexts"] = contexts
        return visible_apps

    def graph(self):
        context = self._state["rezContexts"][self._state["appRequest"]]
        graph_str = context.graph(as_dot=True)

        tempdir = tempfile.mkdtemp()
        fname = os.path.join(tempdir, "graph.png")

        try:
            rez.save_graph(graph_str, fname)
            pixmap = QtGui.QPixmap(fname)

        except IOError:
            self.error("GraphViz not found")
            return QtGui.QPixmap()

        finally:
            # Don't need this no more
            shutil.rmtree(tempdir)

        return pixmap


class Command(QtCore.QObject):
    stdout = QtCore.Signal(str)
    stderr = QtCore.Signal(str)
    killed = QtCore.Signal()

    error = QtCore.Signal(Exception)

    def __str__(self):
        return "Command('%s')" % self.cmd

    def __init__(self,
                 context,
                 command,
                 package,
                 overrides=None,
                 disabled=None,
                 detached=True,
                 environ=None,
                 parent=None):
        super(Command, self).__init__(parent)

        self.overrides = overrides or {}  # unused
        self.disabled = disabled or {}  # unused
        self.environ = environ or {}

        self.context = context
        self.app = package
        self.popen = None
        self.detached = detached

        # `cmd` rather than `command`, to distinguish
        # between class and argument
        self.cmd = command

        self._running = False

        # Launching may take a moment, and there's no need
        # for the user to wait around for that to happen.
        thread = threading.Thread(target=self._execute)
        thread.daemon = True

        self.thread = thread

    @property
    def pid(self):
        if self.popen.poll is None:
            return self.popen.pid

    def execute(self):
        self.thread.start()

    def _execute(self):
        startupinfo = None
        no_console = hasattr(allzparkconfig, "__noconsole__")

        # Windows-only
        # Prevent additional windows from appearing when running
        # Allzpark without a console, e.g. via pythonw.exe.
        if no_console and hasattr(subprocess, "STARTUPINFO"):
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        kwargs = {
            "command": self.cmd,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE,
            "parent_environ": None,
            "startupinfo": startupinfo
        }

        context = self.context

        if self.environ:
            # Inject user environment
            #
            # NOTE: Rez takes precendence on environment, so a user
            # cannot edit the environment in such a way that packages break.
            # However it also means it cannot edit variables also edited
            # by a package. Win some lose some
            kwargs["parent_environ"] = dict(os.environ, **self.environ)

        try:
            self.popen = context.execute_shell(**kwargs)
        except Exception as e:
            return self.error.emit(e)

        for target in (self.listen_on_stdout,
                       self.listen_on_stderr):
            thread = threading.Thread(target=target)
            thread.daemon = True
            thread.start()

    def is_running(self):
        # Normally, you'd be able to determine whether a Popen instance was
        # still running by querying Popen.poll() == None, but Rez may or may
        # not use `Popen(shell=True)` which throws this mechanism off. Instead,
        # we'll let an open pipe to STDOUT determine whether or not a process
        # is currently running.
        return self._running

    def listen_on_stdout(self):
        self._running = True
        for line in iter(self.popen.stdout.readline, ""):
            self.stdout.emit(line.rstrip())
        self._running = False
        self.killed.emit()

    def listen_on_stderr(self):
        for line in iter(self.popen.stderr.readline, ""):
            self.stderr.emit(line.rstrip())



import os
import time

# Debugging, measure start-up time
# NOTE: Handle this prior to importing anything
if os.getenv("ALLZPARK_STARTTIME"):

    try:
        t0 = float(os.getenv("ALLZPARK_STARTTIME"))
        t1 = time.time()

    except ValueError:
        raise ValueError(
            "ALLZPARK_STARTTIME must be in format time.time()"
        )

    duration = t1 - t0
    print("shell to python: %.2f s" % duration)


import sys
from . import cli
sys.exit(cli.main())



from .vendor.Qt import QtWidgets, QtCore


class Package(QtWidgets.QStyledItemDelegate):
    def __init__(self, ctrl, parent=None):
        super(Package, self).__init__(parent)
        self._ctrl = ctrl

    def createEditor(self, parent, option, index):
        if index.column() != 1:
            return

        editor = QtWidgets.QComboBox(parent)

        return editor

    def setEditorData(self, editor, index):
        model = index.model()
        options = model.data(index, "versions")
        default = index.data(QtCore.Qt.DisplayRole)

        editor.addItems(options)
        editor.setCurrentIndex(options.index(default))

    def setModelData(self, editor, model, index):
        model = index.model()
        package = model.data(index, "name")
        options = model.data(index, "versions")
        default = model.data(index, "default")
        version = options[editor.currentIndex()]

        if not version or version == default:
            return

        self._ctrl.patch("%s==%s" % (package, version))



"""
    transitions.extensions.factory
    ------------------------------

    This module contains the definitions of classes which combine the functionality of transitions'
    extension modules. These classes can be accessed by names as well as through a static convenience
    factory object.
"""

from ..core import Machine

from .nesting import HierarchicalMachine, NestedTransition, NestedEvent
from .locking import LockedMachine, LockedEvent
from .diagrams import GraphMachine, TransitionGraphSupport, NestedGraph


class MachineFactory(object):
    """
        Convenience factory for machine class retrieval.
    """

    # get one of the predefined classes which fulfill the criteria
    @staticmethod
    def get_predefined(graph=False, nested=False, locked=False):
        """ A function to retrieve machine classes by required functionality.
        Args:
            graph (bool): Whether the returned class should contain graph support.
            nested: Whether the returned machine class should support nested states.
            locked: Whether the returned class should facilitate locks for threadsafety.

        Returns (class): A machine class with the specified features.
        """
        return _CLASS_MAP[(graph, nested, locked)]


class NestedGraphTransition(TransitionGraphSupport, NestedTransition):
    """
        A transition type to be used with (subclasses of) `HierarchicalGraphMachine` and
        `LockedHierarchicalGraphMachine`.
    """
    pass


class LockedNestedEvent(LockedEvent, NestedEvent):
    """
        An event type to be used with (subclasses of) `LockedHierarchicalMachine`
        and `LockedHierarchicalGraphMachine`.
    """
    pass


class HierarchicalGraphMachine(GraphMachine, HierarchicalMachine):
    """
        A hierarchical state machine with graph support.
    """

    transition_cls = NestedGraphTransition
    graph_cls = NestedGraph


class LockedHierarchicalMachine(LockedMachine, HierarchicalMachine):
    """
        A threadsafe hierarchical machine.
    """

    event_cls = LockedNestedEvent


class LockedGraphMachine(GraphMachine, LockedMachine):
    """
        A threadsafe machine with graph support.
    """
    pass


class LockedHierarchicalGraphMachine(GraphMachine, LockedMachine, HierarchicalMachine):
    """
        A threadsafe hiearchical machine with graph support.
    """

    transition_cls = NestedGraphTransition
    event_cls = LockedNestedEvent
    graph_cls = NestedGraph


# 3d tuple (graph, nested, locked)
_CLASS_MAP = {
    (False, False, False): Machine,
    (False, False, True): LockedMachine,
    (False, True, False): HierarchicalMachine,
    (False, True, True): LockedHierarchicalMachine,
    (True, False, False): GraphMachine,
    (True, False, True): LockedGraphMachine,
    (True, True, False): HierarchicalGraphMachine,
    (True, True, True): LockedHierarchicalGraphMachine
}
